// Copyright (C) 2004, Brian Enigma <enigma at netninja.com>
// This file is part of BarcodeProcessor.
//
// BarcodeProcessor is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// BarcodeProcessor is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Foobar; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
package org.ninjasoft.barcodeprocessor;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.Hashtable;
import javax.imageio.*;
import javax.swing.*;

/**
 * Class to take a barcode image and turn it into understandable data structure.
 * At present, this ONLY understands wide/narrow barcodes.  It cannot process
 * barcodes with multiple widths (for example, Code 128).
 * @author enigma
 */
public class BarcodeImageProcessor {
    public String imageFilename = null;
    private int wideElementPixels = 0;
    private int narrowElementPixels = 0;
    private BufferedImage inputImage = null;
    private WritableRaster inputRaster = null;
    private int[] barColor = new int[] {0x00, 0x00, 0x00};
    private int[] spaceColor = new int[] {0xff, 0xff, 0xff};
    private boolean useFuzzyPixelMatching = false;
    private boolean debug = false;
    
    private void loadImage() throws IOException {
        inputImage = this.inputImage = ImageIO.read(new File(imageFilename));
        inputRaster = inputImage.getRaster();
    }
    
    // Look for the longest and shortest runs of pixels to guess the wide and
    // narrow widths.
    public void guessWidths() throws IOException {
        int longestRun = 0;
        int narrowestRun = 1024;
        int currentRun = 1;
        boolean recordingBar = false;
        int[] pixel = new int[3];
        // Load the image if not already loaded
        if (inputImage == null)
            loadImage();
        // Check whether we are strting with a bar or a space
        int y = inputImage.getHeight() / 2;
        pixel = inputRaster.getPixel(0, y, pixel);
        if (pixelsEqual(pixel, barColor))
            recordingBar = true;
        else if (pixelsEqual(pixel, spaceColor))
            recordingBar = false;
        else
            pixelError(pixel, 0, y);
        if (this.debug)
        	    System.out.println("First pixel is a " + (recordingBar ? "bar" : "space"));
        // Run through the rest of the image
        for (int x=1; x<inputImage.getWidth(); x++) {
            boolean currentIsBar = false;
            // Check if the current traversed pixel is a bar or space
            pixel = inputRaster.getPixel(x, y, pixel);
            // Sanity check (this is a bit inefficient to call pixelsEqual so many times, 
            // but that is not as big a deal for such a small image as a barcode)
            if (pixelsEqual(pixel, barColor) && pixelsEqual(pixel, spaceColor))
                throw new RuntimeException("Cannot determine if pixel is bar or space (bar and space shades are ill-defined and too close together)");
            if (pixelsEqual(pixel, barColor))
                currentIsBar = true;
            else if (pixelsEqual(pixel, spaceColor))
                currentIsBar = false;
            else
                pixelError(pixel, x, y);
            //System.out.println("(" + x + "," + y + ") is (" + pixel[0] + "," + pixel[1] + "," + pixel[2] + ")");
            // Check if we are continuing the current run of bar or space, or whether the color has flipflopped
            if (recordingBar == currentIsBar) {
                currentRun++;
            } else {
                // Check if the current run beats any records
                if (this.debug)
                	    System.out.println("Found " + (recordingBar ? "bar" : "space") + " of width " + currentRun);
                longestRun = Math.max(longestRun, currentRun);
                narrowestRun = Math.min(narrowestRun, currentRun);
                recordingBar = !recordingBar;
                currentRun = 1;
            }
        }
        // Check end of run (implied color flip-flop becase we reached the edge of the image)
        longestRun = Math.max(longestRun, currentRun);
        narrowestRun = Math.min(narrowestRun, currentRun);
        // Save it to the member variables
        System.out.println("longestRun is " + longestRun + " pixel" + (longestRun != 1 ? "s" : ""));
        System.out.println("narrowestRun is " + narrowestRun + " pixel" + (narrowestRun != 1 ? "s" : ""));
        this.wideElementPixels = longestRun;
        this.narrowElementPixels = narrowestRun;
    }
    
    // Look for the longest and shortest runs of pixels to guess the wide and
    // narrow widths.
    public BarcodeData translate() throws IOException {
        BarcodeData result = new BarcodeData();
        int currentRun = 1;
        boolean recordingBar = false;
        int[] pixel = new int[3];
        // Load the image if not already loaded
        if (inputImage == null)
            loadImage();
        // Check whether we are strting with a bar or a space
        int y = inputImage.getHeight() / 2;
        pixel = inputRaster.getPixel(0, y, pixel);
        if (pixelsEqual(pixel, barColor))
            recordingBar = true;
        else if (pixelsEqual(pixel, spaceColor))
            recordingBar = false;
        else
            pixelError(pixel, 0, y);
        // Run through the rest of the image
        for (int x=1; x<inputImage.getWidth(); x++) {
            boolean currentIsBar = false;
            // Check if the current traversed pixel is a bar or space
            pixel = inputRaster.getPixel(x, y, pixel);
            // Sanity check (this is a bit inefficient to call pixelsEqual so many times, 
            // but that is not as big a deal for such a small image as a barcode)
            if (pixelsEqual(pixel, barColor) && pixelsEqual(pixel, spaceColor))
                throw new RuntimeException("Cannot determine if pixel is bar or space (bar and space shades are ill-defined and too close together)");
            if (pixelsEqual(pixel, barColor))
                currentIsBar = true;
            else if (pixelsEqual(pixel, spaceColor))
                currentIsBar = false;
            else
                pixelError(pixel, x, y);
            // Check if we are continuing the current run of bar or space, or whether the color has flipflopped
            if (recordingBar == currentIsBar) {
                currentRun++;
            } else {
                // Check if the current run beats any records
                if (currentRun == this.narrowElementPixels) {
                    if (recordingBar)
                        result.addNarrowBar();
                    else
                        result.addNarrowSpace();
                } else if (currentRun == this.wideElementPixels) {
                	    if (recordingBar)
                        result.addWideBar();
                    else
                        result.addWideSpace();
                } else {
                    throw new RuntimeException("Found " + (recordingBar ? "bar" : "space") + " of width " + currentRun + " which does not match wide or narrow length");
                }
                recordingBar = !recordingBar;
                currentRun = 1;
            }
        }
        // Check end of run (implied color flip-flop becase we reached the edge of the image)
        // Check if the current run beats any records
        if (currentRun == this.narrowElementPixels) {
            if (recordingBar)
                result.addNarrowBar();
            else
                result.addNarrowSpace();
        } else if (currentRun == this.wideElementPixels) {
                if (recordingBar)
                result.addWideBar();
            else
                result.addWideSpace();
        } else {
            throw new RuntimeException("Found " + (recordingBar ? "bar" : "space") + " of width " + currentRun + " which does not match wide or narrow length");
        }
        return result;
    }

    private void pixelError(int[] pixel, int x, int y) throws RuntimeException {
        throw new RuntimeException("Found color (0x" +
                Integer.toString(pixel[0], 16) + ",0x" +
                Integer.toString(pixel[1], 16) + ",0x" +
                Integer.toString(pixel[2], 16) + ") " +
                "at (" + x + "," + y + ") that was neither bar nor space");
    }
    
    /**
     * Check if the color of two pixels are equal (within 5 shades of color)
     * @param a first pixel
     * @param b second pixel
     * @return colors are equal?
     */
    private boolean pixelsEqual(int[] a, int[] b) {
        if (this.useFuzzyPixelMatching)
            return (Math.abs(a[0] - b[0]) < 5) &&
                   (Math.abs(a[1] - b[1]) < 5) &&
                   (Math.abs(a[2] - b[2]) < 5);
        else
            return (a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]);
    }
    
    // GETTERS AND SETTERS /////////////////////////////////////////////////////
    
	public String getImageFilename() {
		return imageFilename;
	}
	public void setImageFilename(String imageFilename) {
		this.imageFilename = imageFilename;
	}
	public int getNarrowElementPixels() {
		return narrowElementPixels;
	}
	public void setNarrowElementPixels(int narrowElementPixels) {
		this.narrowElementPixels = narrowElementPixels;
	}
	public int getWideElementPixels() {
		return wideElementPixels;
	}
	public void setWideElementPixels(int wideElementPixels) {
		this.wideElementPixels = wideElementPixels;
	}
    public void setBarColor(int red, int green, int blue) {
        if ((red < 0) || (red > 255)) throw new RuntimeException("Red must be between 0 and 255");
        if ((green < 0) || (green > 255)) throw new RuntimeException("Green must be between 0 and 255");
        if ((blue < 0) || (blue > 255)) throw new RuntimeException("Blue must be between 0 and 255");
        this.barColor = new int[] {red, green, blue};
    }
    public void setSpaceColor(int red, int green, int blue) {
        if ((red < 0) || (red > 255)) throw new RuntimeException("Red must be between 0 and 255");
        if ((green < 0) || (green > 255)) throw new RuntimeException("Green must be between 0 and 255");
        if ((blue < 0) || (blue > 255)) throw new RuntimeException("Blue must be between 0 and 255");
        this.spaceColor = new int[] {red, green, blue};
    }
    public void setUseFuzzyPixelMatching(boolean b) {
        this.useFuzzyPixelMatching = b;
    }
    public boolean getUseFuzzyPixelMatching() {
        return this.useFuzzyPixelMatching;
    }
	public void setDebug(boolean debug) {
		this.debug = debug;
	}
}
